# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import Dict, List, Tuple, Any
import ast

from pandas import DataFrame, Series, isna
from edfi_google_classroom_extractor.mapping.constants import SOURCE_SYSTEM


# States returned by API
CREATED_STATE = "CREATED"
NEW_STATE = "NEW"
RECLAIMED_STATE = "RECLAIMED_BY_STUDENT"
TURNED_IN_STATE = "TURNED_IN"
RETURNED_STATE = "RETURNED"

# States derived from API "late" flag
LATE_STATE = "LATE"
MISSING_STATE = "MISSING"


def derive_state(submission_row: Series) -> str:
    """
    Takes a Pandas row of API assign submission data and returns the submission
    state for that row based on the API provided state and late flag.

    Parameters
    ----------
    pandas_row: Any
        is a row of assignment submission data

    Returns
    -------
    str
        Submission state for the row of data
    """
    api_state: str = submission_row["state"]
    if "late" not in submission_row:
        return api_state

    if isna(submission_row["late"]):
        return api_state

    if isinstance(submission_row["late"], bool) and submission_row["late"] is False:
        return api_state

    if (
        isinstance(submission_row["late"], str)
        and submission_row["late"].lower() != "true"
    ):
        return api_state

    if api_state == TURNED_IN_STATE:
        return LATE_STATE

    if (
        api_state == CREATED_STATE
        or api_state == NEW_STATE
        or api_state == RECLAIMED_STATE
    ):
        return MISSING_STATE

    return api_state


def _get_submission_datetime(submission_row: Series) -> str:
    if submission_row["state"] != TURNED_IN_STATE and submission_row["state"] != RETURNED_STATE:
        return ""

    status_history: List[dict] = ast.literal_eval(submission_row["submissionHistory"])
    # this submission history also has a history of grades, but for this purpose, we only care
    # about stateHistory so we filter by stateHistory
    status_history = [
        status_item["stateHistory"]
        for status_item in status_history
        if "stateHistory" in status_item.keys()
    ]
    status_history = sorted(
        status_history, key=lambda k: k["stateTimestamp"], reverse=True
    )

    for status_record in status_history:
        status_is_turned_in: bool = status_record["state"] == TURNED_IN_STATE

        if status_is_turned_in:
            return status_record["stateTimestamp"]

    return ""


def submissions_to_assignment_submissions_dfs(
    submissions_df: DataFrame,
) -> Dict[Tuple[str, str], DataFrame]:
    """
    Convert a Submission API DataFrame to a Dict of AssignmentSubmission UDM DataFrames
    grouped by source system section id/assignment id tuple pairs

    Parameters
    ----------
    submissions_df: DataFrame
        is a Submission API DataFrame

    Returns
    -------
    Dict[Tuple[str, str], DataFrame]
        LMS UDM AssignmentSubmission DataFrames grouped by
            source system section id/assignment id tuple pairs

    Notes
    -----
    AssignmentSubmission DataFrame columns are:
        AssignmentSourceSystemIdentifier: A unique numeric identifier assigned to the assignment
        EarnedPoints: The points earned for the submission
        Grade: The grade received for the submission
        SourceSystem: The system code or name providing the AssignmentSubmission data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to
            an AssignmentSubmission by the source system
        SubmissionStatus: The status of the submission in relation to the late acceptance policy
        SubmissionDateTime: The date and time of the assignment submission
        LMSUserSourceSystemIdentifier: A unique numeric identifier assigned to the user
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
    """
    assert "courseId" in submissions_df.columns
    assert "courseWorkId" in submissions_df.columns
    assert "id" in submissions_df.columns
    assert "userId" in submissions_df.columns
    assert "creationTime" in submissions_df.columns
    assert "updateTime" in submissions_df.columns
    assert "state" in submissions_df.columns
    assert "assignedGrade" in submissions_df.columns

    submissions_df["SourceSystemIdentifier"] = submissions_df[
        ["courseId", "courseWorkId", "id"]
    ].agg("-".join, axis=1)

    submissions_df["AssignmentSourceSystemIdentifier"] = submissions_df[
        ["courseId", "courseWorkId"]
    ].agg("-".join, axis=1)

    submissions_df["Grade"] = submissions_df["assignedGrade"]
    submissions_df["SubmissionDateTime"] = submissions_df.apply(
        _get_submission_datetime,
        axis=1,
    )
    submissions_df["SubmissionStatus"] = submissions_df.apply(derive_state, axis=1)

    assignment_submissions_df: DataFrame = submissions_df[
        [
            "SourceSystemIdentifier",
            "AssignmentSourceSystemIdentifier",
            "Grade",
            "SubmissionDateTime",
            "assignedGrade",
            "userId",
            "courseId",
            "creationTime",
            "updateTime",
            "SubmissionStatus",
            "CreateDate",
            "LastModifiedDate",
        ]
    ]

    assignment_submissions_df = assignment_submissions_df.rename(
        columns={
            "assignedGrade": "EarnedPoints",
            "userId": "LMSUserSourceSystemIdentifier",
            "courseId": "SourceSystemSectionIdentifier",
            "creationTime": "SourceCreateDate",
            "updateTime": "SourceLastModifiedDate",
        }
    )

    assignment_submissions_df["SourceSystem"] = SOURCE_SYSTEM

    # group by section id and assignment id as a Dict of DataFrames
    result: Dict[
        Any, DataFrame
    ] = dict(  # Any because Pylance doesn't believe Tuple[str, str]
        tuple(
            assignment_submissions_df.groupby(
                [
                    "SourceSystemSectionIdentifier",
                    "AssignmentSourceSystemIdentifier",
                ]
            )
        )
    )

    # no longer need group by column
    for grouped_df in result.values():
        grouped_df.drop(columns=["SourceSystemSectionIdentifier"], inplace=True)

    return result



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import Dict
from pandas import DataFrame, concat
from edfi_google_classroom_extractor.mapping.constants import SOURCE_SYSTEM

ENROLLMENT_STATUS_ACTIVE = "Active"


def _students_or_teachers_to_user_section_associations_df(
    students_or_teachers_df: DataFrame,
) -> DataFrame:
    """
    Convert a Student or Teacher API DataFrame to a UserSectionAssociation DataFrame

    Parameters
    ----------
    students_or_teachers_df: DataFrame
        is a Students or Teachers API DataFrame

    Returns
    -------
    DataFrame
        a UserSectionAssociation DataFrame based on the given Students or Teachers API DataFrame

    Notes
    -----
    UserSectionAssociation DataFrame columns are:
        EnrollmentStatus: The status of the user section association. E.g., Active,
            Inactive, Withdrawn
        LMSSectionSourceSystemIdentifier: A unique numeric identifier assigned to the section
        SourceSystem: The system code or name providing the user data
        SourceSystemIdentifier: A unique number or alphanumeric code
            assigned to a user by the source system
        LMSUserSourceSystemIdentifier: A unique numeric identifier assigned to the user
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """
    assert "userId" in students_or_teachers_df.columns
    assert "courseId" in students_or_teachers_df.columns

    user_section_associations_df: DataFrame = students_or_teachers_df[
        [
            "userId",
            "courseId",
            "CreateDate",
            "LastModifiedDate",
        ]
    ]

    user_section_associations_df = user_section_associations_df.rename(
        columns={
            "userId": "LMSUserSourceSystemIdentifier",
            "courseId": "LMSSectionSourceSystemIdentifier",
        }
    )

    user_section_associations_df[
        "SourceSystemIdentifier"
    ] = user_section_associations_df[["LMSUserSourceSystemIdentifier", "LMSSectionSourceSystemIdentifier"]].agg(
        "-".join, axis=1
    )

    user_section_associations_df["SourceSystem"] = SOURCE_SYSTEM
    user_section_associations_df["EnrollmentStatus"] = ENROLLMENT_STATUS_ACTIVE
    user_section_associations_df["SourceCreateDate"] = ""  # No create date available from API
    user_section_associations_df[
        "SourceLastModifiedDate"
    ] = ""  # No modified date available from API
    return user_section_associations_df


def students_and_teachers_to_user_section_associations_dfs(
    students_df: DataFrame, teachers_df: DataFrame
) -> Dict[str, DataFrame]:
    """
    Convert Student and Teacher API DataFrames to a Dict of
    UserSectionAssociation UDM DataFrames grouped by source system
    section id

    Parameters
    ----------
    students_df: DataFrame
        is a Students API DataFrame
    teachers_df: DataFrame
        is a Students API DataFrame

    Returns
    -------
    Dict[str, DataFrame]
        LMS UDM UserSectionAssociation DataFrames grouped by
            source system section id

    Notes
    -----
    UserSectionAssociation DataFrame columns are:
        EnrollmentStatus: The status of the user section association. E.g., Active,
            Inactive, Withdrawn
        LMSSectionSourceSystemIdentifier: A unique numeric identifier assigned to the section
        SourceSystem: The system code or name providing the user data
        SourceSystemIdentifier: A unique number or alphanumeric code
            assigned to a user by the source system
        LMSUserSourceSystemIdentifier: A unique numeric identifier assigned to the user
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """
    assert "userId" in students_df.columns
    assert "courseId" in students_df.columns
    assert "userId" in teachers_df.columns
    assert "courseId" in teachers_df.columns

    associations_from_students_df: DataFrame = (
        _students_or_teachers_to_user_section_associations_df(
            students_df,
        )
    )
    associations_from_teachers_df: DataFrame = (
        _students_or_teachers_to_user_section_associations_df(
            teachers_df,
        )
    )

    user_section_associations_df = concat(
        [associations_from_students_df, associations_from_teachers_df],
        ignore_index=True,
        sort=False,
    )

    # group by section id as a Dict of DataFrames
    result: Dict[str, DataFrame] = dict(
        tuple(
            user_section_associations_df.groupby(
                [
                    "LMSSectionSourceSystemIdentifier",
                ]
            )
        )
    )

    return result



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List, Dict, cast
from pandas import DataFrame, Series, json_normalize
import sqlalchemy
from googleapiclient.discovery import Resource
from edfi_google_classroom_extractor.api.api_caller import call_api, ResourceType
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)


COURSES_RESOURCE_NAME = "Courses"
EDFI_LMS_PREFIX = "EdFiLMS."

logger = logging.getLogger(__name__)


def request_course_aliases(resource: Resource, course_id: str) -> List[Dict[str, str]]:
    """
    Fetch CourseAlias API data for the given course id

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    course_id: str
        a Google Classroom course id

    Returns
    -------
    List[Dict[str, str]]
        a list of Google Classroom CourseAliases for the given course id,
            see https://developers.google.com/classroom/reference/rest/v1/courses.aliases
    """
    return call_api(
        cast(ResourceType, resource).courses().aliases().list,
        {"courseId": course_id},
        "aliases",
    )


def _alias_names_for_scope(aliases: List[str], scope: str) -> List[str]:
    """
    Returns a list of alias names from a list of aliases with the given scope.
    Aliases from the API are scoped with a prefix character followed by a colon
    and then the alias name. For example "d:alias_name_1".

    Parameters
    ----------
    aliases: List[str]
        a list of course alias strings from the API
    scope: str
        the scope to filter by

    Returns
    -------
    List[str]
        a list of alias names (scope omitted) filtered by the given scope
    """
    result: List[str] = []
    try:
        result = [
            alias.split(":")[1] for alias in aliases if alias.split(":")[0] == scope
        ]
    except (ValueError, IndexError):
        pass
    return result


def _find_first_with_edfilms_prefix(alias_names: List[str]) -> str:
    """
    Returns the first alias name in the list prefixed with the
    Ed-Fi LMS indicator, with the prefix removed, or an empty string if none are found.

    Parameters
    ----------
    alias_names: List[str]
        a list of alias names

    Returns
    -------
    str
        The first alias name in the list prefixed with the
        Ed-Fi LMS indicator, with the prefix removed, or an empty string if none are found.
    """
    return next(
        (
            alias_name.removeprefix(EDFI_LMS_PREFIX)  # type: ignore
            for alias_name in alias_names
            if alias_name.startswith(EDFI_LMS_PREFIX)
        ),
        "",
    )


def _select_alias(aliases: List[str]) -> str:
    """
    Selects the most relevant course alias from a list of course aliases returned by the API

    Parameters
    ----------
    aliases: List[str]
        a list of course alias strings from the API, including the "d:" (domain scope) prefix

    Returns
    -------
    str
        a (possibly empty) string representing the most relevant alias for the course.

    Notes
    -----
    How to find the correct SISSectionIdentifier from CourseAlias results:

    If there are no aliases:
    SISSectionIdentifier = ""

    Else if there is exactly one alias:
    SISSectionIdentifier = alias

    Else if there is a domain alias prefixed with "EdFiLMS.":
    SISSectionIdentifier = first domain alias prefixed with "EdFiLMS.", prefix excluded

    Else if there are domain aliases:
    SISSectionIdentifier = first domain alias

    Else:
    SISSectionIdentifier = ""
    """
    if len(aliases) == 0:
        return ""

    if len(aliases) == 1:
        try:
            _, alias_name = aliases[0].split(":")
            return alias_name.removeprefix(EDFI_LMS_PREFIX)  # type: ignore
        except ValueError:
            return ""

    domain_alias_names: List[str] = _alias_names_for_scope(aliases, "d")
    edfilms_domain_alias = _find_first_with_edfilms_prefix(domain_alias_names)
    if edfilms_domain_alias != "":
        return edfilms_domain_alias

    if len(domain_alias_names) > 0:
        return domain_alias_names[0]

    return ""


def _derive_alias(resource: Resource, row: Series) -> str:
    """
    Derives a single alias for a course dataframe row, from a fetch of CourseAliases

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    row: Series
        a Pandas Series row for a single Google Classroom course

    Returns
    -------
    str
        a (possibly empty) string representing the most relevant alias for the course.
    """
    course_aliases: List[Dict[str, str]] = request_course_aliases(resource, row["id"])
    alias_strings: List[str] = [
        course_alias["alias"] for course_alias in course_aliases
    ]
    return _select_alias(alias_strings)


def request_courses(resource: Resource) -> List[Dict[str, str]]:
    """
    Fetch Course API data for all courses and return a list of course data

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource

    Returns
    -------
    List[Dict[str, str]]
        a list of Google Classroom Course resources,
            see https://developers.google.com/classroom/reference/rest/v1/courses
    """
    return call_api(
        cast(ResourceType, resource).courses().list,
        {"courseStates": "ACTIVE"},
        "courses",
    )


def request_latest_courses_as_df(resource: Resource) -> DataFrame:
    """
    Fetch Course API data for all courses and return a Courses API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource

    Returns
    -------
    DataFrame
        a Courses API DataFrame with the fetched data

    Notes
    -----
    DataFrame columns are:
        id: Identifier for this course assigned by Classroom
        name: Name of the course
        section: Section of the course
        descriptionHeading: Optional heading for the description
        description: Optional description
        room: Optional room location
        ownerId: The identifier of the owner of a course
        creationTime: Creation time of the course
        updateTime: Time of the most recent update to this course
        enrollmentCode: Enrollment code to use when joining this course
        courseState: State of the course
        alternateLink: Absolute link to this course in the Classroom web UI
        teacherGroupEmail: The email address of a Google group containing all teachers of the course
        courseGroupEmail: The email address of a Google group containing all members of the course
        guardiansEnabled: Whether or not guardian notifications are enabled for this course
        calendarId: The Calendar ID for a calendar that all course members can see
        alias: A course alias selected from the available aliases for the course
    """

    logger.info("Pulling course data")
    courses: List[Dict[str, str]] = request_courses(resource)
    courses_df: DataFrame = json_normalize(courses)
    courses_df["alias"] = courses_df.apply(
        lambda row: _derive_alias(resource, row), axis=1
    )
    return courses_df


def request_all_courses_as_df(
    resource: Resource, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Fetch Course API data for all courses and return a Courses API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Courses API DataFrame with the current and previously fetched data

    Notes
    -----
    DataFrame columns are:
        id: Identifier for this course assigned by Classroom
        name: Name of the course
        section: Section of the course
        descriptionHeading: Optional heading for the description
        description: Optional description
        room: Optional room location
        ownerId: The identifier of the owner of a course
        creationTime: Creation time of the course
        updateTime: Time of the most recent update to this course
        enrollmentCode: Enrollment code to use when joining this course
        courseState: State of the course
        alternateLink: Absolute link to this course in the Classroom web UI
        teacherGroupEmail: The email address of a Google group containing all teachers of the course
        courseGroupEmail: The email address of a Google group containing all members of the course
        teacherFolder.id: The identifier of the teacher folder
        teacherFolder.title: The identifier of the teacher folder,
        teacherFolder.alternateLink: Absolute link to the teacher folder in the Classroom web UI
        guardiansEnabled: Whether or not guardian notifications are enabled for this course
        calendarId: The Calendar ID for a calendar that all course members can see
        alias: A course alias selected from the available aliases for the course
        CreateDate: Date this record was created by the extractor
        LastModifiedDate: Date this record was last updated by the extractor
    """

    courses_df = request_latest_courses_as_df(resource)
    courses_df = _sync_without_cleanup(courses_df, sync_db)
    cleanup_after_sync(COURSES_RESOURCE_NAME, sync_db)

    return courses_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Courses API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id"],
        resource_name=COURSES_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from pandas import DataFrame, concat
from edfi_google_classroom_extractor.mapping.constants import SOURCE_SYSTEM


STUDENT_USER_ROLE = "Student"
TEACHER_USER_ROLE = "Teacher"


def _students_or_teachers_to_users_df(
    students_or_teachers_df: DataFrame, lms_udm_user_role: str
) -> DataFrame:
    """
    Convert a Students or Teachers API DataFrame to a LMSUsers DataFrame

    Parameters
    ----------
    students_or_teachers_df: DataFrame
        is a Students or Teachers API DataFrame
    lms_udm_user_role
        is the LMS UDM user role as a string

    Returns
    -------
    DataFrame
        a LMSUsers DataFrame based on the given Students or Teachers API DataFrame

    Notes
    -----
    DataFrame columns are:
        EmailAddress: The primary e-mail address for the user
        LocalUserIdentifier: The user identifier assigned by a school or district
        Name: The full name of the user
        SISUserIdentifier: The user identifier defined in the Student Information System (SIS)
        SourceSystem: The system code or name providing the user data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a user by the source system
        UserRole: The role assigned to the user
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """
    assert "userId" in students_or_teachers_df.columns
    assert "profile.name.fullName" in students_or_teachers_df.columns
    assert "profile.emailAddress" in students_or_teachers_df.columns

    result: DataFrame = students_or_teachers_df[
        [
            "userId",
            "profile.name.fullName",
            "profile.emailAddress",
            "CreateDate",
            "LastModifiedDate",
        ]
    ]
    result = result.rename(
        columns={
            "userId": "SourceSystemIdentifier",
            "profile.name.fullName": "Name",
            "profile.emailAddress": "EmailAddress",
        }
    )

    # Student records are per-course, so there may be duplicates
    result.drop_duplicates(inplace=True)
    result["SourceSystem"] = SOURCE_SYSTEM
    result["UserRole"] = lms_udm_user_role

    result["LocalUserIdentifier"] = ""  # No local id available from API
    result["SISUserIdentifier"] = ""  # No SIS id available from API
    result["SourceCreateDate"] = ""  # No create date available from API
    result["SourceLastModifiedDate"] = ""  # No modified date available from API

    return result


def students_and_teachers_to_users_df(
    students_df: DataFrame, teachers_df: DataFrame
) -> DataFrame:
    """
    Convert Students and Teachers API DataFrames to an LMS UDM DataFrame

    Parameters
    ----------
    students_df: DataFrame
        is a Students API DataFrame
    teachers_df: DataFrame
        is a Teachers API DataFrame

    Returns
    -------
    DataFrame
        a LMSUsers DataFrame based on the given Students and Teachers API DataFrames

    Notes
    -----
    DataFrame columns are:
        EmailAddress: The primary e-mail address for the user
        LocalUserIdentifier: The user identifier assigned by a school or district
        Name: The full name of the user
        SISUserIdentifier: The user identifier defined in the Student Information System (SIS)
        SourceSystem: The system code or name providing the user data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a user by the source system
        UserRole: The role assigned to the user
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """
    assert "userId" in students_df.columns
    assert "profile.name.fullName" in students_df.columns
    assert "profile.emailAddress" in students_df.columns

    assert "userId" in teachers_df.columns
    assert "profile.name.fullName" in teachers_df.columns
    assert "profile.emailAddress" in teachers_df.columns

    users_from_students_df: DataFrame = _students_or_teachers_to_users_df(
        students_df, STUDENT_USER_ROLE
    )
    users_from_teachers_df: DataFrame = _students_or_teachers_to_users_df(
        teachers_df, TEACHER_USER_ROLE
    )

    return concat(
        [users_from_students_df, users_from_teachers_df], ignore_index=True, sort=False
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List, Dict, Optional, cast
from pandas import DataFrame, json_normalize
import sqlalchemy
from googleapiclient.discovery import Resource
from edfi_google_classroom_extractor.api.api_caller import call_api, ResourceType
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)

STUDENTS_RESOURCE_NAME = "Students"

logger = logging.getLogger(__name__)


def request_students(
    resource: Optional[Resource], course_id: str
) -> List[Dict[str, str]]:
    """
    Fetch Students API data for a course and return a list of student data

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    course_ids: str
        a Google Classroom course id as a string

    Returns
    -------
    List[Dict[str, str]]
        a list of Google Classroom Student resources,
            see https://developers.google.com/classroom/reference/rest/v1/courses.students
    """

    if resource is None:
        return []

    return call_api(
        cast(ResourceType, resource).courses().students().list,
        {"courseId": course_id},
        "students",
    )


def request_latest_students_as_df(
    resource: Optional[Resource], course_ids: List[str]
) -> DataFrame:
    """
    Fetch Students API data for a range of courses and return a Students API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource or None
    course_ids: List[str]
        a list of Google Classroom course ids as a string array

    Returns
    -------
    DataFrame
        a Students API DataFrame with the fetched data

    Notes
    -----
    DataFrame columns are:
        courseId: Identifier of the course
        userId: Identifier of the user
        profile.id: Identifier of the user
        profile.name.givenName: The user's first name
        profile.name.familyName: The user's last name
        profile.name.fullName: The user's full name formed by concatenating the first and last name values
        profile.emailAddress: Email address of the user
    """

    logger.info("Pulling student data")
    students: List[Dict[str, str]] = []
    for course_id in course_ids:
        students.extend(request_students(resource, course_id))

    return json_normalize(students).astype("string")


def request_all_students_as_df(
    resource: Optional[Resource],
    course_ids: List[str],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch Students API data for a range of courses and return a Students API DataFrame
    with current and previously fetched data

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource or None
    course_ids: List[str]
        a list of Google Classroom course ids as a string array
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Students API DataFrame with the current and previously fetched data

    Notes
    -----
    DataFrame columns are:
        courseId: Identifier of the course
        userId: Identifier of the user
        profile.id: Identifier of the user
        profile.name.givenName: The user's first name
        profile.name.familyName: The user's last name
        profile.name.fullName: The user's full name formed by concatenating the first and last name values
        profile.emailAddress: Email address of the user
    """

    students_df: DataFrame = request_latest_students_as_df(resource, course_ids)
    students_df = _sync_without_cleanup(students_df, sync_db)
    cleanup_after_sync(STUDENTS_RESOURCE_NAME, sync_db)

    return students_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a Students API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["userId", "courseId"],
        resource_name=STUDENTS_RESOURCE_NAME,
        sync_db=sync_db,
    )




from dataclasses import dataclass
from typing import List

from configargparse import ArgParser

from . import constants


@dataclass
class MainArguments:
    """
    Container for holding arguments parsed at the command line.
    """
    classroom_account: str
    log_level: str
    output_directory: str
    usage_start_date: str
    usage_end_date: str
    sync_database_directory: str
    extract_activities: bool = False
    extract_assignments: bool = False
    extract_attendance: bool = False
    extract_grades: bool = False


def parse_main_arguments(args_in: List[str]) -> MainArguments:
    """
    Configures the command-line interface.

    Parameters
    ----------
    args_in : list of str
        Full argument list from the command line.

    Returns
    -------
    arguments  : MainArguments
        A populated `MainArguments` object.
    """

    parser = ArgParser()

    parser.add(  # type: ignore
        "-a",
        "--classroom-account",
        required=True,
        help="The email address of the Google Classroom admin account.",
        type=str,
        default="",
        env_var="CLASSROOM_ACCOUNT",
    )

    parser.add(  # type: ignore
        "-l",
        "--log-level",
        required=False,
        help="The log level for the tool.",
        choices=constants.LOG_LEVELS,
        type=str,
        default="INFO",
        env_var="LOG_LEVEL",
    )

    parser.add(  # type: ignore
        "-o",
        "--output-directory",
        required=False,
        help="The output directory for the generated csv files.",
        type=str,
        default="data/",
        env_var="OUTPUT_DIRECTORY",
    )

    parser.add(  # type: ignore
        "-s",
        "--usage-start-date",
        required=False,
        help="Start date for usage data pull in yyyy-mm-dd format.",
        type=str,
        default="",
        env_var="START_DATE",
    )

    parser.add(  # type: ignore
        "-e",
        "--usage-end-date",
        required=False,
        help="End date for usage data pull in yyyy-mm-dd format.",
        type=str,
        default="",
        env_var="END_DATE",
    )

    parser.add(  # type: ignore
        "-d",
        "--sync-database-directory",
        required=False,
        help="The directory for the sync database.",
        type=str,
        default="data",
        env_var="SYNC_DATABASE_DIRECTORY",
    )

    parser.add(  # type: ignore
        "-f",
        "--feature",
        required=False,
        help="Features to include.",
        type=str,
        nargs='*',
        choices=constants.VALID_FEATURES,
        default=[],
        env_var="FEATURE",
    )

    args_parsed = parser.parse_args(args_in)

    assert isinstance(
        args_parsed.output_directory, str
    ), "The specified `classroom-account` is not valid."

    arguments = MainArguments(
        classroom_account=args_parsed.classroom_account,
        log_level=args_parsed.log_level,
        output_directory=args_parsed.output_directory,
        usage_start_date=args_parsed.usage_start_date,
        usage_end_date=args_parsed.usage_end_date,
        sync_database_directory=args_parsed.sync_database_directory,
        extract_activities=constants.Features.Activities in args_parsed.feature,
        extract_assignments=constants.Features.Assignments in args_parsed.feature,
        extract_attendance=constants.Features.Attendance in args_parsed.feature,
        extract_grades=constants.Features.Grades in args_parsed.feature,
    )

    return arguments



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import Dict
from datetime import datetime
from pandas import DataFrame
from edfi_google_classroom_extractor.mapping.constants import SOURCE_SYSTEM


def coursework_to_assignments_dfs(
    coursework_df: DataFrame,
) -> Dict[str, DataFrame]:
    """
    Convert a Coursework API DataFrame to a Dict of Assignment UDM DataFrames
    grouped by source system section id pairs

    Parameters
    ----------
    coursework_df: DataFrame
        is a Coursework API DataFrame

    Returns
    -------
    Dict[str, DataFrame]
        LMS UDM Assignment DataFrames grouped by source system section id

    Notes
    -----
    Assignment DataFrame columns are:
        AssignmentCategory: The category or type of assignment
        AssignmentDescription: The assignment description
        DueDateTime: The date and time the assignment is due
        EndDateTime: The end date and time for the assignment
        LMSSectionSourceSystemIdentifier: A unique numeric identifier assigned
            to the section
        MaxPoints: The maximum number of points a student may receive
        SourceSystem: The system code or name providing the assignment data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to
            an assignment by the source system
        StartDateTime: The start date and time for the assignment
        Title: The assignment title or name
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
    """
    assert "courseId" in coursework_df.columns
    assert "id" in coursework_df.columns
    assert "workType" in coursework_df.columns
    assert "description" in coursework_df.columns
    assert "scheduledTime" in coursework_df.columns
    assert "maxPoints" in coursework_df.columns
    assert "title" in coursework_df.columns

    if {
        "dueDate.year",
        "dueDate.month",
        "dueDate.day",
        "dueTime.hours",
        "dueTime.minutes",
    }.issubset(coursework_df.columns):
        due_date_df: DataFrame = coursework_df[
            [
                "dueDate.year",
                "dueDate.month",
                "dueDate.day",
                "dueTime.hours",
                "dueTime.minutes",
            ]
        ]
        coursework_df["DueDateTime"] = due_date_df.apply(
            lambda date_element: ""
            if date_element.isnull().any()
            else datetime(*date_element.astype(float).astype(int)),
            axis=1,
        )
    else:
        coursework_df["DueDateTime"] = ""

    coursework_df["SourceSystemIdentifier"] = coursework_df[["courseId", "id"]].agg(
        "-".join, axis=1
    )

    assignments_df: DataFrame = coursework_df[
        [
            "courseId",
            "workType",
            "description",
            "scheduledTime",
            "maxPoints",
            "title",
            "creationTime",
            "updateTime",
            "DueDateTime",
            "SourceSystemIdentifier",
            "CreateDate",
            "LastModifiedDate",
        ]
    ]

    assignments_df = assignments_df.rename(
        columns={
            "courseId": "LMSSectionSourceSystemIdentifier",
            "workType": "AssignmentCategory",
            "description": "AssignmentDescription",
            "scheduledTime": "StartDateTime",
            "maxPoints": "MaxPoints",
            "title": "Title",
            "creationTime": "SourceCreateDate",
            "updateTime": "SourceLastModifiedDate",
        }
    )

    assignments_df["SubmissionType"] = assignments_df["AssignmentCategory"]
    assignments_df["SourceSystem"] = SOURCE_SYSTEM
    assignments_df["EndDateTime"] = ""  # No EndDateTime available from API

    # group by section id as a Dict of DataFrames
    result: Dict[str, DataFrame] = dict(
        tuple(assignments_df.groupby(["LMSSectionSourceSystemIdentifier"]))
    )

    return result



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from datetime import datetime, timedelta
import logging
from typing import List, Dict, Optional, Any, cast
from dateutil.parser import parse as date_parse
from pandas import DataFrame, json_normalize, read_sql, date_range
from sqlalchemy.exc import OperationalError
import sqlalchemy
from googleapiclient.discovery import Resource
from .api_caller import call_api, ResourceType

logger = logging.getLogger(__name__)


def request_usage(resource: Optional[Resource], date: str) -> List[Dict[str, str]]:

    if resource is None:
        return []

    return call_api(
        cast(ResourceType, resource).userUsageReport().get,
        {
            "userKey": "all",
            "date": date,
            "parameters": "classroom:timestamp_last_interaction,classroom:num_posts_created,accounts:timestamp_last_login",
        },
        "usageReports",
    )


def last_sync_date(sync_db: sqlalchemy.engine.base.Engine) -> Optional[datetime]:
    with sync_db.connect() as con:
        try:
            usage_df = read_sql("SELECT asOfDate FROM Usage", con)
            if usage_df["asOfDate"].count() == 0:
                return None
            return date_parse(usage_df["asOfDate"].max())  # type: ignore
        except OperationalError:
            logger.debug("No Usage table yet")
            return None


def start_date(sync_db: sqlalchemy.engine.base.Engine, env_start_date: str) -> datetime:
    last_date: Optional[datetime] = last_sync_date(sync_db)
    if last_date is not None:
        return last_date + timedelta(days=1)
    if env_start_date == "":
        return datetime.today()
    return date_parse(env_start_date)


def end_date(env_end_date: str) -> datetime:
    if env_end_date == "":
        return datetime.today()
    return date_parse(env_end_date)


def request_latest_usage_as_df(
    resource: Optional[Resource], start: datetime, end: datetime
) -> DataFrame:
    logger.info("Pulling usage data")

    if end < start:
        logger.info("Usage data end time is before start time.")

    reports: List[Any] = []
    for date in date_range(start=start, end=end):
        reports.extend(request_usage(resource, date.strftime("%Y-%m-%d")))

    usage: List[Dict[str, str]] = []
    for response in reports:
        row: Dict[str, str] = {}
        row["email"] = response.get("entity").get("userEmail")
        row["asOfDate"] = response.get("date")
        row["importDate"] = datetime.today().strftime("%Y-%m-%d")

        for parameter in response.get("parameters"):
            if parameter.get("name") == "classroom:num_posts_created":
                row["numberOfPosts"] = parameter.get("intValue")

            if parameter.get("name") == "classroom:last_interaction_time":
                row["lastInteractionTime"] = parameter.get("datetimeValue")

            if parameter.get("name") == "accounts:last_login_time":
                row["lastLoginTime"] = parameter.get("datetimeValue")
        usage.append(row)

    usage_df: DataFrame = json_normalize(usage)
    if usage_df.empty:
        return usage_df

    usage_df = usage_df.astype(
        {
            "email": "string",
            "asOfDate": "datetime64",
            "importDate": "datetime64",
            "numberOfPosts": "int32",
            "lastInteractionTime": "datetime64",
            "lastLoginTime": "datetime64",
        }
    )

    usage_df["name"] = usage_df["email"].str.split("@").str[0]
    usage_df["monthDay"] = usage_df["asOfDate"].dt.strftime("%m/%d")
    usage_df["nameDate"] = usage_df["name"] + " " + usage_df["monthDay"]

    return usage_df


def request_all_usage_as_df(
    resource: Optional[Resource],
    sync_db: sqlalchemy.engine.base.Engine,
    env_start_date: str,
    env_end_date: str,
) -> DataFrame:
    usage_df: DataFrame = request_latest_usage_as_df(
        resource, start_date(sync_db, env_start_date), end_date(env_end_date)
    )
    if usage_df.empty:
        return usage_df

    usage_df.to_sql("Usage", sync_db, if_exists="append", index=False, chunksize=500)
    # remove duplicates - leave only the most recent
    with sync_db.connect() as con:
        con.execute(
            "DELETE from Usage "
            "WHERE rowid not in (select max(rowid) "
            "FROM Usage "
            "GROUP BY email, asOfDate)"
        )

    return usage_df



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List, Dict, Optional, cast
from pandas import DataFrame, json_normalize
import sqlalchemy
from googleapiclient.discovery import Resource
from .api_caller import call_api, ResourceType
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)

REQUIRED_COLUMNS = [
    "courseId",
    "id",
    "title",
    "description",
    "state",
    "alternateLink",
    "creationTime",
    "updateTime",
    "maxPoints",
    "workType",
    "submissionModificationMode",
    "assigneeMode",
    "creatorUserId",
    "dueDate.year",
    "dueDate.month",
    "dueDate.day",
    "dueTime.hours",
    "dueTime.minutes",
    "scheduledTime",
    "topicId",
]

ASSIGNMENTS_RESOURCE_NAME = "Assignmments"

logger = logging.getLogger(__name__)


def request_coursework(
    resource: Optional[Resource], course_id: str
) -> List[Dict[str, str]]:
    """
    Fetch Coursework API data for all coursework for a course,
    and return a list of coursework data

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource

    Returns
    -------
    List[Dict[str, str]]
        a list of Google Classroom Coursework resources,
            see https://developers.google.com/classroom/reference/rest/v1/courses.courseWork
    """

    if resource is None:
        return []

    return call_api(
        cast(ResourceType, resource).courses().courseWork().list,
        {"courseId": course_id},
        "courseWork",
    )


def request_latest_coursework_as_df(
    resource: Optional[Resource], course_ids: List[str]
) -> DataFrame:
    """
    Fetch Coursework API data for the given courses
        and return a Coursework API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    course_ids: List[str]
        a list of course ids to retrieve coursework for

    Returns
    -------
    DataFrame
        a Coursework API DataFrame with the fetched data

    Notes
    -----
    DataFrame columns are:
        courseId: Identifier of the course
        id: Classroom-assigned identifier of this course work, unique per course
        title: Title of this course work
        description: Optional description of this course work
        state: Status of this course work
        alternateLink: Absolute link to this course work in the Classroom web UI
        creationTime: Timestamp when this course work was created
        updateTime: Timestamp of the most recent change to this course work
        maxPoints: Maximum grade for this course work
        workType: Type of this course work
        submissionModificationMode: Setting to determine when students are allowed
            to modify submissions
        assigneeMode: Assignee mode of the coursework
        creatorUserId: Identifier for the user that created the coursework
        dueDate.year: Optional year, in UTC, that submissions are due
        dueDate.month: Optional month, in UTC, that submissions are due
        dueDate.day: Optional day, in UTC, that submissions are due
        dueTime.hours: Optional hour, in UTC, that submissions are due
        dueTime.minutes: Optional minute, in UTC, that submissions are due
        scheduledTime: Optional timestamp when this course work is scheduled to be published
        topicId: Identifier for the topic that this coursework is associated with
    """

    logger.info("Pulling coursework data")
    coursework: List[Dict[str, str]] = []
    for course_id in course_ids:
        coursework.extend(request_coursework(resource, course_id))

    json_df: DataFrame = json_normalize(coursework).astype("string")
    return json_df.reindex(
        json_df.columns.union(REQUIRED_COLUMNS, sort=False), axis=1, fill_value=""  # type: ignore
    )


def request_all_coursework_as_df(
    resource: Optional[Resource],
    course_ids: List[str],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch Coursework API data for all courses and return a Coursework API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    course_ids: List[str]
        a list of course ids to retrieve coursework for
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a Coursework API DataFrame with the current and previously fetched data

    Notes
    -----
    DataFrame columns are:
        courseId: Identifier of the course
        id: Classroom-assigned identifier of this course work, unique per course
        title: Title of this course work
        description: Optional description of this course work
        state: Status of this course work
        alternateLink: Absolute link to this course work in the Classroom web UI
        creationTime: Timestamp when this course work was created
        updateTime: Timestamp of the most recent change to this course work
        maxPoints: Maximum grade for this course work
        workType: Type of this course work
        submissionModificationMode: Setting to determine when students are allowed
            to modify submissions
        assigneeMode: Assignee mode of the coursework
        creatorUserId: Identifier for the user that created the coursework
        dueDate.year: Optional year, in UTC, that submissions are due
        dueDate.month: Optional month, in UTC, that submissions are due
        dueDate.day: Optional day, in UTC, that submissions are due
        dueTime.hours: Optional hour, in UTC, that submissions are due
        dueTime.minutes: Optional minute, in UTC, that submissions are due
        scheduledTime: Optional timestamp when this course work is scheduled to be published
        topicId: Identifier for the topic that this coursework is associated with
        CreateDate: Date this record was created by the extractor
        LastModifiedDate: Date this record was last updated by the extractor
    """

    coursework_df: DataFrame = request_latest_coursework_as_df(resource, course_ids)
    coursework_df = _sync_without_cleanup(coursework_df, sync_db)
    cleanup_after_sync(ASSIGNMENTS_RESOURCE_NAME, sync_db)

    return coursework_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a courseworks API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections


    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id", "courseId"],
        resource_name=ASSIGNMENTS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from collections import namedtuple
import socket
import logging
import os
from typing import List, Dict, Optional, Callable, cast
from requests import RequestException
from opnieuw import retry
from tail_recursive import tail_recursive
from googleapiclient.errors import Error as GoogleApiError

ResourceType = namedtuple("ValidSdkFunction", ["courses", "userUsageReport"])

MAX_TOTAL_CALLS = int(os.getenv("REQUEST_RETRY_COUNT") or 4)
RETRY_WINDOW_AFTER_FIRST_CALL_IN_SECONDS = int(
    os.environ.get("REQUEST_RETRY_TIMEOUT_SECONDS") or 60
)

logger = logging.getLogger(__name__)


@retry(
    retry_on_exceptions=(IOError, RequestException, GoogleApiError, socket.timeout, socket.error),
    max_calls_total=MAX_TOTAL_CALLS,
    retry_window_after_first_call_in_seconds=RETRY_WINDOW_AFTER_FIRST_CALL_IN_SECONDS,
)
def _execute(executable_resource):
    """
    Invoke a get/list Google Classroom SDK function,
    retrying if there are errors.

    Parameters
    ----------
    executable_resource: function
        is the get/list Google Classroom SDK function to call

    Returns
    -------
    object
        a Google Classroom SDK response object

    Raises
    ------
    IOError
        if there is an IOError after retrying
    RequestException
        if there is a RequestException after retrying
    """
    assert hasattr(executable_resource, "execute")
    return executable_resource.execute()


@tail_recursive
def _call_api_recursive(
    resource_method: Callable,
    resource_parameters: Dict[str, str],
    response_property: str,
    results: Optional[List[Dict[str, str]]] = None,
) -> List[Dict[str, str]]:
    """
    Call a Google Classroom/Admin SDK API

    Parameters
    ----------
    resource_method: function
        is the get/list SDK function to call
    resource_parameters: dict
        is the parameters for get/list
    response_property: string
        is the property in the API response we want
    results: list
        is the list of dicts of the API response accumulated across pages

    Returns
    -------
    list
        a list of dicts of the API response property requested
    """

    current_results: List[Dict[str, str]] = [] if results is None else results
    response = _execute(resource_method(**resource_parameters))
    try:
        current_results.extend(response.get(response_property, []))
    except (IOError, RequestException, GoogleApiError):
        logger.exception("Error during API call after retries.  Will try to continue.")
        return current_results

    next_page_token = response.get("nextPageToken", None)
    if not next_page_token:
        return current_results
    resource_parameters["pageToken"] = next_page_token
    return cast(
        List[Dict[str, str]],
        _call_api_recursive.tail_call(
            resource_method, resource_parameters, response_property, current_results
        ),
    )


def call_api(
    resource_method: Callable,
    resource_parameters: Dict[str, str],
    response_property: str,
    results: Optional[List[Dict[str, str]]] = None,
) -> List[Dict[str, str]]:
    """
    Call a Google Classroom/Admin SDK API

    Parameters
    ----------
    resource_method: function
        is the get/list SDK function to call
    resource_parameters: dict
        is the parameters for get/list
    response_property: string
        is the property in the API response we want
    results: list
        is the list of dicts of the API response accumulated across pages

    Returns
    -------
    list
        a list of dicts of the API response property requested
    """

    return _call_api_recursive(resource_method, resource_parameters, response_property, results)  # type: ignore



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import json
from typing import Dict
from pandas import DataFrame, concat, Series
from edfi_google_classroom_extractor.mapping.constants import SOURCE_SYSTEM

ACTIVITY_TYPE_STATE = "Submission State Change"
ACTIVITY_TYPE_GRADE = "Submission Grade Change"


def submissions_to_user_submission_activities_dfs(
    submissions_df: DataFrame,
) -> Dict[str, DataFrame]:
    """
    Convert a Submission API DataFrame to a Dict of UserActivity
    UDM DataFrames grouped by source system section id.

    Parameters
    ----------
    submissions_df: DataFrame
        is a Submission API DataFrame

    Returns
    -------
    Dict[str, DataFrame] LMS UDM UserActivity DataFrames
        grouped by source system section id

    Notes
    -----
    UserActivity DataFrame columns are:
        ActivityDateTime: The date/time the activity occurred
        ActivityStatus: The activity status
        ActivityTimeInMinutes: The total activity time in minutes
        ActivityType: The type of activity, here "Submission" or "Grade"
        AssignmentIdentifier: A unique numeric identifier assigned to the assignment
        Content: Content associated with the activity
        LMSSectionIdentifier: A unique numeric identifier assigned to the section
        SourceSystem: The system code or name providing the user activity data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a
            user activity by the source system
        LMSUserIdentifier: A unique numeric identifier assigned to the user
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """
    assert "submissionHistory" in submissions_df.columns
    assert "id" in submissions_df.columns
    assert "courseId" in submissions_df.columns
    assert "courseWorkId" in submissions_df.columns

    # convert json-like submissionHistory string to list of dicts
    submissions_df["submissionHistory"] = submissions_df["submissionHistory"].apply(lambda json_like: json.loads(json_like.replace("'", '"')))
    submissions_df["AssignmentIdentifier"] = submissions_df[
        ["courseId", "courseWorkId"]
    ].agg("-".join, axis=1)

    submissions_df = submissions_df[["id", "courseId", "courseWorkId", "submissionHistory", "AssignmentIdentifier", "CreateDate", "LastModifiedDate"]]

    # explode submissionHistory lists into rows with other columns duplicated
    history_df = submissions_df.explode(column="submissionHistory")  # type: ignore

    # expand submissionHistory dicts (stateHistory and gradeHistory) into their own columns
    history_df = history_df["submissionHistory"].apply(Series).merge(history_df, left_index=True, right_index=True, how='outer')
    history_df.drop(columns=["submissionHistory"], inplace=True)

    # expand stateHistory (can assume exists, should always have at least one "CREATED" entry)
    user_submission_df = concat([history_df, history_df["stateHistory"].apply(Series)], axis=1)
    user_submission_df.dropna(subset=["stateHistory"], inplace=True)

    # enrich stateHistory
    user_submission_df["SourceSystemIdentifier"] = "S-" + user_submission_df[
        ["courseId", "courseWorkId", "id", "stateTimestamp"]
    ].agg("-".join, axis=1)

    user_submission_df = user_submission_df[
        [
            "SourceSystemIdentifier",
            "AssignmentIdentifier",
            "stateTimestamp",
            "state",
            "courseId",
            "actorUserId",
            "CreateDate",
            "LastModifiedDate"
        ]
    ]

    user_submission_df = user_submission_df.rename(
        columns={
            "stateTimestamp": "ActivityDateTime",
            "state": "ActivityStatus",
            "courseId": "LMSSectionIdentifier",
            "actorUserId": "LMSUserIdentifier",
        }
    )

    user_submission_df["ActivityType"] = ACTIVITY_TYPE_STATE

    # expand gradeHistory if exists
    if "gradeHistory" in history_df:
        grade_history_df = concat([history_df, history_df["gradeHistory"].apply(Series)], axis=1)
        grade_history_df.dropna(subset=["gradeHistory"], inplace=True)

        # enrich gradeHistory
        grade_history_df["SourceSystemIdentifier"] = "G-" + grade_history_df[
            ["courseId", "courseWorkId", "id", "gradeTimestamp"]
        ].agg("-".join, axis=1)

        grade_history_df = grade_history_df[
            [
                "SourceSystemIdentifier",
                "AssignmentIdentifier",
                "gradeTimestamp",
                "gradeChangeType",
                "courseId",
                "actorUserId",
                "CreateDate",
                "LastModifiedDate"
            ]
        ]

        grade_history_df = grade_history_df.rename(
            columns={
                "gradeTimestamp": "ActivityDateTime",
                "gradeChangeType": "ActivityStatus",
                "courseId": "LMSSectionIdentifier",
                "actorUserId": "LMSUserIdentifier",
            }
        )

        grade_history_df["ActivityType"] = ACTIVITY_TYPE_GRADE

        # combine with stateHistory
        user_submission_df = user_submission_df.append(grade_history_df)

    # teacher actions can show up on student histories and vice-versa
    user_submission_df.drop_duplicates(subset=["SourceSystemIdentifier"], inplace=True)

    # finish with common columns
    user_submission_df["ActivityTimeInMinutes"] = ""
    user_submission_df["Content"] = ""
    user_submission_df["SourceSystem"] = SOURCE_SYSTEM
    user_submission_df["SourceCreateDate"] = ""  # No create date available from API
    user_submission_df["SourceLastModifiedDate"] = ""  # No modified date available from API

    # group by section id as a Dict of DataFrames
    result: Dict[str, DataFrame] = dict(
        tuple(user_submission_df.groupby(["LMSSectionIdentifier"]))
    )

    return result



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
from typing import List, Dict, Optional, cast
from pandas import DataFrame, json_normalize
import sqlalchemy
from googleapiclient.discovery import Resource
from .api_caller import call_api, ResourceType
from edfi_lms_extractor_lib.api.resource_sync import (
    cleanup_after_sync,
    sync_to_db_without_cleanup,
)

REQUIRED_COLUMNS = [
    "courseId",
    "courseWorkId",
    "id",
    "userId",
    "creationTime",
    "updateTime",
    "state",
    "late",
    "draftGrade",
    "assignedGrade",
    "alternateLink",
    "courseWorkType",
    "associatedWithDeveloper",
    "submissionHistory",
]

SUBMISSIONS_RESOURCE_NAME = "StudentSubmissions"

logger = logging.getLogger(__name__)


def request_submissions(
    resource: Optional[Resource], course_id: str
) -> List[Dict[str, str]]:
    """
    Fetch Student Submissions API data for all submissions for a coursework,
    and return a list of submissions data

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource

    Returns
    -------
    List[Dict[str, str]]
        a list of Google Classroom StudentSubmissions resources,
            see https://developers.google.com/classroom/reference/rest/v1/courses.courseWork.studentSubmissions
    """

    if resource is None:
        return []

    return call_api(
        cast(ResourceType, resource).courses().courseWork().studentSubmissions().list,
        {"courseId": course_id, "courseWorkId": "-"},
        "studentSubmissions",
    )


def request_latest_submissions_as_df(
    resource: Optional[Resource], course_ids: List[str]
) -> DataFrame:
    """
    Fetch StudentSubmissions API data for the given coursework
        and return a StudentSubmissions API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    course_ids: List[str]
        a list of course ids to retrieve coursework for

    Returns
    -------
    DataFrame
        a Coursework API DataFrame with the fetched data

    Notes
    -----
    DataFrame columns are:
        courseId: Identifier of the course
        courseWorkId: Identifier for the course work this corresponds to
        id: Classroom-assigned Identifier for the student submission, unique
        per coursework
        userId: Identifier for the student that owns this submission
        creationTime: Creation time of this submission
        updateTime: Last update time of this submission
        title: Title of this course work
        description: Optional description of this course work
        state: Status of this submission
        late: Whether this submission is late
        draftGrade: Optional pending grade. If unset, no grade was set. Decimal
            values are allowed.
        assignedGrade: Optional grade. If unset, no grade was set. Decimal
            values are allowed.
        alternateLink: Absolute link to this course work in the Classroom web UI
        courseWorkType: Type of course work this submission is for
        associatedWithDeveloper: Whether this student submission is associated with the
            Developer Console project making the request
        submissionHistory: The history of the submission as JSON
    """

    logger.info("Pulling student submission data")
    submissions: List[Dict[str, str]] = []
    for course_id in course_ids:
        submissions.extend(request_submissions(resource, course_id))

    json_df: DataFrame = json_normalize(submissions).astype("string")
    return json_df.reindex(
        json_df.columns.union(REQUIRED_COLUMNS, sort=False), axis=1, fill_value=""  # type: ignore
    )


def request_all_submissions_as_df(
    resource: Optional[Resource],
    course_ids: List[str],
    sync_db: sqlalchemy.engine.base.Engine,
) -> DataFrame:
    """
    Fetch StudentSubmissions API data for the given coursework
        and return a StudentSubmissions API DataFrame

    Parameters
    ----------
    resource: Optional[Resource]
        a Google Classroom SDK Resource
    course_ids: List[str]
        a list of course ids to retrieve coursework for

    Returns
    -------
    DataFrame
        a Coursework API DataFrame with the fetched data

    Notes
    -----
    DataFrame columns are:
        courseId: Identifier of the course
        courseWorkId: Identifier for the course work this corresponds to
        id: Classroom-assigned Identifier for the student submission, unique
        per coursework
        userId: Identifier for the student that owns this submission
        creationTime: Creation time of this submission
        updateTime: Last update time of this submission
        state: Status of this submission
        late: Whether this submission is late
        draftGrade: Optional pending grade. If unset, no grade was set. Decimal
            values are allowed.
        assignedGrade: Optional grade. If unset, no grade was set. Decimal
            values are allowed.
        alternateLink: Absolute link to this course work in the Classroom web UI
        courseWorkType: Type of course work this submission is for
        associatedWithDeveloper: Whether this student submission is associated with the
            Developer Console project making the request
        submissionHistory: The history of the submission as JSON
        CreateDate: Date this record was created by the extractor
        LastModifiedDate: Date this record was last updated by the extractor
    """

    submissions_df: DataFrame = request_latest_submissions_as_df(
        resource, course_ids
    )

    submissions_df = _sync_without_cleanup(submissions_df, sync_db)
    cleanup_after_sync(SUBMISSIONS_RESOURCE_NAME, sync_db)

    return submissions_df


def _sync_without_cleanup(
    resource_df: DataFrame, sync_db: sqlalchemy.engine.base.Engine
) -> DataFrame:
    """
    Take fetched API data and sync with database. Creates tables when necessary,
    but ok if temporary tables are there to start. Doesn't delete temporary tables when finished.

    Parameters
    ----------
    resource_df: DataFrame
        a courseworks API DataFrame with the current fetched data which
        will be mutated, adding Hash and CreateDate/LastModifiedDate
    sync_db: sqlalchemy.engine.base.Engine
        an Engine instance for creating database connections

    Returns
    -------
    DataFrame
        a DataFrame with current fetched data and reconciled CreateDate/LastModifiedDate
    """
    return sync_to_db_without_cleanup(
        resource_df=resource_df,
        identity_columns=["id", "courseId", "courseWorkId"],
        resource_name=SUBMISSIONS_RESOURCE_NAME,
        sync_db=sync_db,
    )



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

import logging
import os

from google.oauth2 import service_account
from sqlalchemy import create_engine
import sqlalchemy


logger = logging.getLogger(__name__)


def _is_running_in_notebook() -> bool:
    """
    Determine whether code is running in a Jupyter Notebook

    Returns
    -------
    bool
        True if code is running in a Jupyter Notebook
    """
    main = __import__("__main__", None, None, fromlist=["__file__"])
    return not hasattr(main, "__file__")


def get_sync_db_engine(sync_database_directory: str) -> sqlalchemy.engine.base.Engine:
    """
    Create a SQL Alchemy Engine for a SQLite file

    Returns
    -------
    sqlalchemy.engine.base.Engine
        a SQL Alchemy Engine
    """
    running_in_notebook: bool = _is_running_in_notebook()
    logger.debug("Running in Jupyter Notebook: %s", running_in_notebook)
    logger.debug("Ensuring database directory at %s", os.path.abspath(sync_database_directory))
    os.makedirs(sync_database_directory, exist_ok=True)
    return create_engine(f"sqlite:///{sync_database_directory}/sync.sqlite")


def get_credentials(classroom_account: str) -> service_account.Credentials:
    """
    Create a Google OAuth Credentials object from a service-account.json file

    Returns
    -------
    Credentials
        a Google OAuth Credentials object
    """
    scopes = [
        "https://www.googleapis.com/auth/admin.directory.orgunit",
        "https://www.googleapis.com/auth/admin.reports.usage.readonly",
        "https://www.googleapis.com/auth/classroom.courses",
        "https://www.googleapis.com/auth/classroom.coursework.students",
        "https://www.googleapis.com/auth/classroom.profile.emails",
        "https://www.googleapis.com/auth/classroom.rosters",
        "https://www.googleapis.com/auth/classroom.student-submissions.students.readonly",
        "https://www.googleapis.com/auth/admin.reports.audit.readonly",
    ]

    filename = (
        "service-account.json"
        if os.path.exists("service-account.json")
        else "../service-account.json"
    )

    return service_account.Credentials.from_service_account_file(
        filename, scopes=scopes, subject=classroom_account
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['edfi_google_classroom_extractor',
 'edfi_google_classroom_extractor.api',
 'edfi_google_classroom_extractor.helpers',
 'edfi_google_classroom_extractor.mapping']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.2.3,<2.0.0',
 'SQLAlchemy>=1.3.19,<2.0.0',
 'edfi-lms-extractor-lib>=1.1.0,<2.0.0',
 'edfi-lms-file-utils>=1.0.0,<2.0.0',
 'errorhandler>=2.0.1,<3.0.0',
 'google-api-python-client>=1.11.0,<2.0.0',
 'google-auth-oauthlib>=0.4.1,<0.5.0',
 'numpy==1.19.3',
 'opnieuw>=1.1.0,<2.0.0',
 'pandas>=1.1.1,<2.0.0',
 'pytest>=6.0,<7.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'tail-recursive>=2.0.0,<3.0.0',
 'xxhash>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'edfi-google-classroom-extractor',
    'version': '1.2.0',
    'description': 'Extract tool for retrieving student data from Google Classroom',
    'long_description': '# Google Classroom Extractor\n\nThis tool retrieves and writes out to CSV students, active sections,\nassignments, and submissions by querying the Google Classroom API. For more\ninformation on the this tool and its output files, please see the main\nrepository [readme](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit).\n\n## Getting Started\n\n1. Download the latest code from [the project\n   homepage](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit) by clicking on\n   the green "CODE" button and choosing an appropriate option. If choosing the\n   Zip option, extract the file contents using your favorite zip tool.\n1. Open a command prompt\\* and change to this file\'s directory (\\* e.g. cmd.exe,\n   PowerShell, bash).\n1. Ensure you have [Python 3.9+ and\n   Poetry](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit#getting-started).\n1. At a command prompt, install all required dependencies:\n\n   ```bash\n   poetry install\n   ```\n\n1. Optional: make a copy of the `.env.example` file, named simply `.env`, and\n   customize the settings as described in the Configuration section below.\n1. Place the service-account.json file described below into\n   the root directory of this project.\n1. Run the extractor one of two ways:\n   * Execute the extractor with minimum command line arguments:\n\n      ```bash\n      poetry run python edfi_google_classroom_extractor -a [admin account email] -f assignments\n      ```\n\n   * Alternately, run with environment variables or `.env` file:\n\n     ```bash\n     poetry run python edfi_google_classroom_extractor\n     ```\n\n   * For detailed help, execute `poetry run python canvas_extractor -h`.\n\n## Configuration\n\n### Module Configuration\n\nApplication configuration is provided through environment variables or command\nline interface (CLI) arguments. CLI arguments take precedence over environment\nvariables. Environment variables can be set the normal way, or by using a\ndedicated [`.env` file](https://pypi.org/project/python-dotenv/) like:\n\n```none\nCLASSROOM_ACCOUNT[<email address of the Google Classroom admin account, required]\nLOG_LEVEL=[Log level, optional]\nOUTPUT_DIRECTORY=[The output directory for the csv files, optional]\nSTART_DATE=[start date for usage data pull in yyyy-mm-dd format, optional]\nEND_DATE=[end date for usage data pull in yyyy-mm-dd format, optional]\n```\n\nSupported parameters:\n\n| Description | Required | Command Line Argument | Environment Variable |\n| ----------- | -------- | --------------------- | -------------------- |\n| The email address of the Google Classroom admin account. | yes | `-a` or `--classroom-account` | CLASSROOM_ACCOUNT |\n| The log level for the tool. ** | no (default: INFO) | `-l` or `--log-level` | LOG_LEVEL |\n| The output directory for the generated csv files. | no (default: [working directory]/data) | `-o` or `--output-directory` | OUTPUT_PATH |\n| Sync database directory | no (default: [working directory]/data) | `-d` or `--sync-database-directory` | SYNC_DATABASE_DIRECTORY |\n| Start date*, yyyy-mm-dd format | no (default: today) | `-s` or `--usage-start-date` | START_DATE |\n| End date*, yyyy-mm-dd format | no (default: today) | `-e` or `--usage-end-date` | END_DATE |\n| Number of retry attempts for failed API calls | no (default: 4) | none | REQUEST_RETRY_COUNT |\n| Timeout window for retry attempts, in seconds | no (default: 60 seconds) | none | REQUEST_RETRY_TIMEOUT_SECONDS |\n| Feature*** | no (default: core, not removable) | `-f` or `--feature` | FEATURE |\n\n\\* _Start Date_ and _End Date_ are used in pulling system activity (usage)\ndata and could span any relevant date range.\n\n\\** Valid values for the optional _log level_:\n\n* DEBUG\n* INFO(default)\n* WARNING\n* ERROR\n* CRITICAL\n\n\\*** When there\'s no specified feature, the extractor will always process Users,\nSections, and Section Associations, which are considered the core feature. Other\nfeatures (can combine two or more):\n\n* assignments (Enables the extraction of assignments and submissions)\n* activities (Enables the extraction of section activities and system\n  activities) - **EXPERIMENTAL**, subject to breaking changes\n* grades (Enables the extraction of grades) - **COMING SOON**\n\nWhen setting features via `.env` file or through environment variable, combine\nfeatures by using a bracketed comma-separate list, e.g. `FEATURE=[activities,\nattendance, assignments, grades]`. To combine features at the command line,\nsimply list them together: `--feature activities, attendance, assignments,\ngrades`.\n\nNote: in order to make the extractor work, you still need to configure your\n`service-account.json` file. To do so, read the next section `API Permissions`\n\n### API Permissions\n\nIn order to extract data, the Google Classroom APIs must be\nenabled, and the application must be granted permission.\n\nA Google Classroom administrator will need to enable both the\nGoogle Classroom API and the Admin SDK.  This can be done\n[here](https://console.developers.google.com/apis/library).\n\nNext, the administrator will need to create a Service Account\nand API key.  This is the account the application will use for\naccess.  This can be done\n[here](https://console.cloud.google.com/iam-admin/serviceaccounts/create).\n\n1. Give the new service account a name like "Ed-Fi Extractor"\n   and click Create.\n1. Grant the service account the "Viewer" role and click `Continue` then\n   Done, skipping step 3: "Grant users access to this service account".\n1. The new service account will be displayed in a table.\n   Click on the three dots for the account and select Manage Keys.\n1. On the next page, click the `Add Key` button, then choose JSON and click\n   `Create` in the dialog box.\n1. A JSON file will be downloaded from your browser, which is the API key.\n   Rename it to `service-account.json`. Save this into the project directory.\n1. Finally, click on the service account to view details and\n   copy the Unique ID field for the next step.\n\nFinally, the administrator will need to specify the scope of\naccess for the service account.  This can be done\n[here](https://admin.google.com/ac/owl/domainwidedelegation).\n\n1. Add a new API client and provide the service account Unique\n   ID (`client_id` in the json file) in the `Client ID` field.\n1. Paste the following scopes into the OAuth scopes field and\n   click `Authorize`:\n\n`https://www.googleapis.com/auth/admin.directory.orgunit,\nhttps://www.googleapis.com/auth/admin.reports.usage.readonly,\nhttps://www.googleapis.com/auth/classroom.courses,\nhttps://www.googleapis.com/auth/classroom.coursework.students,\nhttps://www.googleapis.com/auth/classroom.profile.emails,\nhttps://www.googleapis.com/auth/classroom.rosters,\nhttps://www.googleapis.com/auth/classroom.student-submissions.students.readonly,\nhttps://www.googleapis.com/auth/admin.reports.audit.readonly`\n\n### Generate LMS UDM CSV Files\n\nTo pull data from Google Classroom and generate csv files, run\n`poetry run python edfi_google_classroom_extractor` from the root\ndirectory of this project. CSV files are output into the\n`data/ed-fi-udm-lms` directory.\n\n### TLS/SSL proxying\n\nUsers on a corporate network that intercepts TLS/SSL traffic will need to have a\ncopy of the corporate root certificate on file, and then add an environment\nvariable pointing to this file: `HTTPLIB2_CA_CERTS=[absolute path to\ncertificate]`. NOTE: this does not load properly through the `.env` file, and\nmust be set as an actual environment variable.\n\n### Logging and Exit Codes\n\nLog statements are written to the standard output. If you wish to capture log\ndetails, then be sure to redirect the output to a file. For example:\n\n```bash\npoetry run python google_classroom_extractor > 2020-12-07-15-43.log\n```\n\nIf any errors occurred during the script run, then there will be a final print\nmessage to the standard error handler as an additional mechanism for calling\nattention to the error: `"A fatal error occurred, please review the log output\nfor more information."`\n\nThe application will exit with status code `1` if there were any log messages at\nthe ERROR or CRITICAL level, otherwise it will exit with status code `0`.\n\n\n## Course Aliases and SIS Section Identifiers\n\nCourse Aliases in Google Classroom are expected to be used to provide a mapping\nfrom a Google Classroom Course to a SIS section using a SIS section identifer.\nTo enable this behavior, create a domain-scoped Course Alias in Google Classroom\nfor each Course with the prefix "EdFiLMS." followed by the SIS section identifier.\nFor example, the domain-scoped Course Alias "d:EdFiLMS.ALG-123" would be used for\nSIS section identifier "ALG-123". Only the first Course Alias found with the\n"EdFiLMS." prefix will be used.\n\nSee\n[course.aliases](https://developers.google.com/classroom/reference/rest/v1/courses.aliases)\nfor more information on Course Aliases.\n\n\n## Developer Operations\n\n1. Style check: `poetry run flake8`\n1. Static typing check: `poetry run mypy .`\n1. Run unit tests: `poetry run pytest`\n1. Run unit tests with code coverage: `poetry run coverage run -m pytest`\n1. View code coverage: `poetry run coverage report`\n\n_Also see\n[build.py](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/docs/build.md)_\nfor use of the build script.\n\n### Visual Studio Code (Optional)\n\nTo work in Visual Studio Code install the Python Extension.\nThen type `Ctrl-Shift-P`, then choose `Python:Select Interpreter`,\nthen choose the environment that includes `.venv` in the name.\n\n## Legal Information\n\nCopyright (c) 2022 Ed-Fi Alliance, LLC and contributors.\n\nLicensed under the [Apache License, Version\n2.0](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/LICENSE) (the\n"License").\n\nUnless required by applicable law or agreed to in writing, software distributed\nunder the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR\nCONDITIONS OF ANY KIND, either express or implied. See the License for the\nspecific language governing permissions and limitations under the License.\n\nSee\n[NOTICES](https://github.com/Ed-Fi-Exchange-OSS/LMS-Toolkit/blob/main/NOTICES.md)\nfor additional copyright and license notifications.\n',
    'author': 'Ed-Fi Alliance, LLC, and contributors',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://techdocs.ed-fi.org/display/EDFITOOLS/LMS+Toolkit',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from datetime import datetime
import logging
from typing import Any, Dict, List
import sys

from googleapiclient.discovery import build, Resource
from google.oauth2 import service_account
from pandas import DataFrame
import sqlalchemy


from edfi_google_classroom_extractor.api.courses import request_all_courses_as_df
from edfi_google_classroom_extractor.api.coursework import request_all_coursework_as_df
from edfi_google_classroom_extractor.api.students import request_all_students_as_df
from edfi_google_classroom_extractor.api.teachers import request_all_teachers_as_df
from edfi_google_classroom_extractor.api.submissions import (
    request_all_submissions_as_df,
)
from edfi_google_classroom_extractor.helpers.arg_parser import MainArguments
from edfi_google_classroom_extractor.config import get_credentials, get_sync_db_engine
from edfi_google_classroom_extractor.mapping.users import (
    students_and_teachers_to_users_df,
)
from edfi_google_classroom_extractor.mapping.user_section_associations import (
    students_and_teachers_to_user_section_associations_dfs,
)
from edfi_google_classroom_extractor.mapping.sections import courses_to_sections_df
from edfi_google_classroom_extractor.mapping.assignments import (
    coursework_to_assignments_dfs,
)
from edfi_google_classroom_extractor.mapping.assignment_submissions import (
    submissions_to_assignment_submissions_dfs,
)
from edfi_google_classroom_extractor.mapping.user_submission_activities import (
    submissions_to_user_submission_activities_dfs,
)
from edfi_lms_extractor_lib.csv_generation.write import (
    write_grades,
    write_users,
    write_sections,
    write_section_associations,
    write_section_activities,
    write_assignments,
    write_assignment_submissions,
    write_system_activities,
)
from edfi_lms_extractor_lib.helpers.decorators import catch_exceptions


logger = logging.getLogger(__name__)
now = datetime.now()
# This variable facilitates temporary storage of output results from one GET
# request that need to be used for creating another GET request.
result_bucket: Dict[str, Any] = {}


def _break_execution(failing_extraction: str) -> None:
    logger.critical(
        f"Unable to continue file generation because the load of {failing_extraction} failed. Please review the log for more information."
    )
    sys.exit(1)


@catch_exceptions
def _get_courses(
    classroom_resource: Resource,
    sync_db: sqlalchemy.engine.base.Engine,
    output_directory: str,
):
    courses_df: DataFrame = request_all_courses_as_df(classroom_resource, sync_db)
    result_bucket["course_ids"] = courses_df["id"].tolist()

    logger.info("Writing LMS UDM Sections to CSV file")
    (sections_df, all_section_ids) = courses_to_sections_df(courses_df)
    result_bucket["section_ids"] = all_section_ids
    write_sections(
        sections_df,
        now,
        output_directory,
    )


@catch_exceptions
def _get_users(
    classroom_resource: Resource,
    sync_db: sqlalchemy.engine.base.Engine,
    output_directory: str,
):
    course_ids: List[str] = result_bucket["course_ids"]

    students = request_all_students_as_df(classroom_resource, course_ids, sync_db)
    teachers = request_all_teachers_as_df(classroom_resource, course_ids, sync_db)
    result_bucket["students_df"] = students
    result_bucket["teachers_df"] = teachers

    logger.info("Writing LMS UDM Users to CSV file")
    write_users(
        students_and_teachers_to_users_df(students, teachers),
        now,
        output_directory,
    )


@catch_exceptions
def _get_section_associations(classroom_resource: Resource, output_directory: str):
    logger.info("Writing LMS UDM UserSectionAssociations to CSV files")

    students_df = result_bucket["students_df"]
    teachers_df = result_bucket["teachers_df"]
    all_section_ids = result_bucket["section_ids"]

    write_section_associations(
        students_and_teachers_to_user_section_associations_dfs(
            students_df, teachers_df
        ),
        all_section_ids,
        now,
        output_directory,
    )


@catch_exceptions
def _get_assignments(
    classroom_resource: Resource,
    sync_db: sqlalchemy.engine.base.Engine,
    output_directory: str,
):
    logger.info("Writing LMS UDM Assignments to CSV files")

    course_ids: List[str] = result_bucket["course_ids"]
    all_section_ids = result_bucket["section_ids"]
    courseworks_df = request_all_coursework_as_df(
        classroom_resource, course_ids, sync_db
    )

    write_assignments(
        coursework_to_assignments_dfs(courseworks_df),
        all_section_ids,
        now,
        output_directory,
    )


@catch_exceptions
def _get_assignment_submissions(
    classroom_resource: Resource,
    sync_db: sqlalchemy.engine.base.Engine,
    output_directory: str,
):
    logger.info("Writing LMS UDM AssignmentSubmissions to CSV files")

    course_ids: List[str] = result_bucket["course_ids"]
    submissions_df = request_all_submissions_as_df(
        classroom_resource, course_ids, sync_db
    )
    result_bucket["submissions_df"] = submissions_df

    write_assignment_submissions(
        submissions_to_assignment_submissions_dfs(submissions_df),
        now,
        output_directory,
    )


@catch_exceptions
def _get_section_activities(output_directory: str):
    logger.info("Writing LMS UDM Section Activities to CSV files")

    submissions_df: DataFrame = result_bucket["submissions_df"]
    all_section_ids = result_bucket["section_ids"]
    write_section_activities(
        submissions_to_user_submission_activities_dfs(submissions_df),
        all_section_ids,
        now,
        output_directory,
    )


@catch_exceptions
def _get_system_activities(output_directory: str):
    logger.info("Writing empty LMS UDM SystemActivities to CSV file")

    write_system_activities(
        DataFrame(),
        now,
        output_directory,
    )


@catch_exceptions
def _get_grades(output_directory: str):
    logger.info("Writing empty LMS UDM Grades to CSV files")

    all_section_ids = result_bucket["section_ids"]
    write_grades(
        dict(),
        all_section_ids,
        now,
        output_directory,
    )


def run(arguments: MainArguments):
    logger.info("Starting Ed-Fi LMS Google Classroom Extractor")
    credentials: service_account.Credentials = get_credentials(
        arguments.classroom_account
    )
    classroom_resource: Resource = build(
        "classroom", "v1", credentials=credentials, cache_discovery=False
    )
    sync_db: sqlalchemy.engine.base.Engine = get_sync_db_engine(
        arguments.sync_database_directory
    )

    succeeded: bool = False

    succeeded = _get_courses(classroom_resource, sync_db, arguments.output_directory)
    if not succeeded:
        _break_execution("Sections")

    succeeded = _get_users(classroom_resource, sync_db, arguments.output_directory)
    if not succeeded:
        _break_execution("Users")

    succeeded = _get_section_associations(classroom_resource, arguments.output_directory)
    if not succeeded:
        _break_execution("Section Associations")

    if arguments.extract_assignments:
        succeeded = _get_assignments(
            classroom_resource, sync_db, arguments.output_directory
        )
        if not succeeded:
            _break_execution("Assignments")

        _get_assignment_submissions(
            classroom_resource, sync_db, arguments.output_directory
        )

    if arguments.extract_activities:
        _get_section_activities(arguments.output_directory)
        _get_system_activities(arguments.output_directory)

    if arguments.extract_grades:
        _get_grades(arguments.output_directory)

    logger.info("Finishing Ed-Fi LMS Google Classroom Extractor")



# SPDX-License-Identifier: Apache-2.0
# Licensed to the Ed-Fi Alliance under one or more agreements.
# The Ed-Fi Alliance licenses this file to you under the Apache License, Version 2.0.
# See the LICENSE and NOTICES files in the project root for more information.

from typing import List, Tuple
from pandas import DataFrame
from edfi_google_classroom_extractor.mapping.constants import SOURCE_SYSTEM


def courses_to_sections_df(courses_df: DataFrame) -> Tuple[DataFrame, List[str]]:
    """
    Convert a Courses API DataFrame to an LMSSections UDM DataFrame

    Parameters
    ----------
    courses_df: DataFrame
        is a Courses API DataFrame

    Returns
    -------
    Tuple[DataFrame, List[str]]
        a LMSSections DataFrame based on the given Courses API DataFrame, and
        a List of all LMSSection section ids

    Notes
    -----
    DataFrame columns are:
        LMSSectionStatus: The section status from the source system
        SISSectionIdentifier: The section identifier defined in the Student Information System
        SectionDescription: The section description
        SourceSystem: The system code or name providing the section data
        SourceSystemIdentifier: A unique number or alphanumeric code assigned to a user by the source system
        Term: The enrollment term for the section
        Title: The section title or name
        SourceCreateDate: Date this record was created in the LMS
        SourceLastModifiedDate: Date this record was last updated in the LMS
        CreateDate: Date this record was created in the extractor
        LastModifiedDate: Date this record was last updated in the extractor
    """
    assert "id" in courses_df.columns
    assert "courseState" in courses_df.columns
    assert "descriptionHeading" in courses_df.columns
    assert "name" in courses_df.columns

    result: DataFrame = courses_df[
        [
            "id",
            "courseState",
            "descriptionHeading",
            "name",
            "creationTime",
            "updateTime",
            "alias",
            "CreateDate",
            "LastModifiedDate",
        ]
    ]
    result = result.rename(
        columns={
            "id": "SourceSystemIdentifier",
            "courseState": "LMSSectionStatus",
            "descriptionHeading": "SectionDescription",
            "name": "Title",
            "creationTime": "SourceCreateDate",
            "updateTime": "SourceLastModifiedDate",
            "alias": "SISSectionIdentifier",
        }
    )

    result["SourceSystem"] = SOURCE_SYSTEM
    result["Term"] = ""  # No term available from API

    return (result, result["SourceSystemIdentifier"].astype("string").tolist())
